/*
 * Created by diablos on 2007-12-02 12:44:23 +0100
 * All Rights Reserved
 */

#import "PaVeImageRep.h"

#import "PaVeOperators.h"

@interface PaVeImageRep (Private)

- (BOOL) setupLua;
- (void) flushGlobals;

- (void) setupLuaRect: (NSRect) r;
- (BOOL) pushObject: obj;
- (void) pushDict: (NSDictionary *) dict;
- (void) pushArray: (NSArray *) array;

@end

@implementation PaVeImageRep

NSString * const PaVeImagePboardType =
  @"PaVeImagePboardType";

+ (void) load
{
  NSDebugLog (@"PaVe: registered image class");
  [NSImageRep registerImageRepClass: [PaVeImageRep class]];
}

+ (BOOL) canInitWithData: (NSData *) d
{
  if ([d length] < 8)
    return NO;

  const char *bytes = (const char *) [d bytes];
  if (bytes[0] != '-' || bytes[1] != '-' || bytes[2] != '!' ||
      bytes[3] != ' ' || bytes[4] != 'P' || bytes[5] != 'a' ||
      bytes[6] != 'V' || bytes[7] != 'e')
    return NO;

  return YES;
}

+ (NSArray *) imageUnfilteredFileTypes
{
  return [NSArray arrayWithObject: @"pave"];
}

+ (NSArray *) imageUnfilteredPasteboardTypes
{
  return [NSArray arrayWithObject: PaVeImagePboardType];
}

- (void) dealloc
{
  TEST_RELEASE (programData);
  TEST_RELEASE (extraArguments);

  if (L != NULL)
    {
      lua_close (L);
      L = NULL;
    }

  [super dealloc];
}

static NSArray *permittedGlobalSymbols = nil;

+ (void) initialize
{
  if (self == [PaVeImageRep class])
    {
      permittedGlobalSymbols = [[NSArray alloc]
        initWithObjects:
        @"arc",
        @"arcn",
        @"arct",
        @"assert",
        @"availablefontfamilies",
        @"availablefonts",
        @"charpath",
        @"clip",
        @"closepath",
        @"collectgarbage",
        @"concat",
        @"coroutine",
        @"currentalpha",
        @"currentcmykcolor",
        @"currentflat",
        @"currentgray",
        @"currenthsbcolor",
        @"currentlinecap",
        @"currentlinejoin",
        @"currentlinewidth",
        @"currentmiterlimit",
        @"currentpoint",
        @"currentrgbcolor",
        @"currentstrokeadjust",
        @"curveto",
        @"eoclip",
        @"eofill",
        @"error",
        @"fill",
        @"findfont",
        @"flattenpath",
        @"fontinfo",
        @"_G",
        @"gcinfo",
        @"getfenv",
        @"getmetatable",
        @"grestore",
        @"gsave",
        @"initclip",
        @"initmatrix",
        @"ipairs",
        @"lineto",
        @"_LOADED",
        @"loadstring",
        @"math",
        @"moveto",
        @"newpath",
        @"newproxy",
        @"next",
        @"os",
        @"pairs",
        @"pathbbox",
        @"pcall",
        @"__pow",
        @"print",
        @"rawequal",
        @"rawget",
        @"rawset",
        @"rcurveto",
        @"rectclip",
        @"rectfill",
        @"rectstroke",
        @"reversepath",
        @"rlineto",
        @"rmoveto",
        @"rotate",
        @"scale",
        @"setalpha",
        @"setcmykcolor",
        @"setdash",
        @"setfenv",
        @"setflat",
        @"setfont",
        @"setgray",
        @"sethalftonephase",
        @"sethsbcolor",
        @"setlinecap",
        @"setlinejoin",
        @"setlinewidth",
        @"setmetatable",
        @"setmiterlimit",
        @"setrgbcolor",
        @"setstrokeadjust",
        @"show",
        @"string",
        @"stroke",
        @"table",
        @"tonumber",
        @"tostring",
        @"translate",
        @"type",
        @"unpack",
        @"_VERSION",
        @"widthofstring",
        @"xpcall",
        nil];
    }
}

enum {
  PaVeVersionMajor = 1,
  PaVeVersionMinor = 0
};

- initWithData: (NSData *) d
{
  if ((self = [self init]) != nil)
    {
      ASSIGN (name, @"(unnamed)");

      programData = [[NSString alloc]
        initWithData: d encoding: NSUTF8StringEncoding];

      if (programData == nil)
        {
          [self release];

          return nil;
        }

      NSMutableDictionary *args = [NSMutableDictionary dictionary];

      // now to parse the PaVe header

      NSArray *lines = [programData componentsSeparatedByString: @"\n"];
      NSString *firstLine = [lines objectAtIndex: 0];

      // first line and version check
      if (![firstLine hasPrefix: @"--! PaVe "])
        {
          NSDebugLog (@"Error parsing PaVe image header: PaVe magic not found");

          [self release];

          return nil;
        }

      int major, minor;

      if (sscanf ([firstLine cString], "--! PaVe %d.%d", &major, &minor) != 2)
        {
          NSDebugLog (@"Error parsing PaVe image header: failed to scan PaVe image version");

          [self release];

          return nil;
        }

      if (major > PaVeVersionMajor || (major == PaVeVersionMajor &&
        minor > PaVeVersionMinor))
        {
          NSDebugLog (@"Error parsing PaVe image header: the file version (%d.%d) is higher "
            @"than what I can handle (%d.%d)", major, minor, PaVeVersionMajor, PaVeVersionMinor);

          [self release];

          return nil;
        }

      int i, n;

      for (i = 1, n = [lines count]; i < n; i++)
        {
          NSString *line = [lines objectAtIndex: i];

          if (![line hasPrefix: @"--! "])
            break;

          NSRange r = [line rangeOfString: @":"];

          if (r.location == NSNotFound)
            break;

          NSString *tagName = [line substringWithRange:
            NSMakeRange (4, r.location - 4)],
                   *tagValue = [line substringFromIndex: r.location + r.length + 1];

          if ([tagName isEqualToString: @"Name"])
            {
              ASSIGN (name, tagValue);
            }
            // unused
/*          else if ([tagName isEqualToString: @"Argument"])
            {
              NSMutableArray *comps = [[[tagValue componentsSeparatedByString:
                @" "] mutableCopy] autorelease];

              [comps removeObject: @""];

              if ([comps count] != 2)
                {
                  NSDebugLog (@"Error parsing PaVe image %@: malformed "
                    @"Argument: header on line %i", name, i);

                  [self release];

                  return nil;
                }

              [args setObject: [comps objectAtIndex: 0]
                       forKey: [comps objectAtIndex: 1]];
            }*/
          else if ([tagName isEqualToString: @"PixelsHigh"])
            {
              [self setPixelsHigh: [tagValue intValue]];
            }
          else if ([tagName isEqualToString: @"PixelsWide"])
            {
              [self setPixelsWide: [tagValue intValue]];
            }
          else if ([tagName isEqualToString: @"Size"])
            {
              double w, h;

              if (sscanf ([tagValue cString], "%lfx%lf", &w, &h) != 2)
                {
                  NSDebugLog (@"Error parsing PaVe image %@ header: malformed "
                    @"Size: header on line %i", name, i);

                  [self release];

                  return nil;
                }
            }
          else if ([tagName isEqualToString: @"Opaque"])
            {
              [self setOpaque: [tagValue boolValue]];
            }
          else if ([tagName isEqualToString: @"Alpha"])
            {
              [self setAlpha: [tagValue boolValue]];
            }
          else
            {
              NSDebugLog (@"Warning parsing PaVe image %@ header: unknown "
                @"header %@ found on line %i", name, tagName, i);
            }
        }

      ASSIGNCOPY (extraArguments, args);
    }

  return self;
}

- (void) setName: (NSString *) aName
{
  ASSIGN (name, aName);
}

- (NSString *) name
{
  return name;
}

- (void) setExtraArguments: (NSDictionary *) dict
{
  ASSIGNCOPY (extraArguments, dict);
}

- (NSDictionary *) extraArguments
{
  return extraArguments;
}

- (void) draw
{
  [self drawAtPoint: NSZeroPoint];
}

- (BOOL) drawAtPoint: (NSPoint) aPoint
{
  NSSize s = [self size];

  return [self drawInRect: NSMakeRect (aPoint.x, aPoint.y, s.width, s.height)];
}

- (BOOL) drawInRect: (NSRect) rect
{
  return [self drawInRect: rect arguments: nil];
}

- (BOOL) drawInRect: (NSRect) r arguments: (NSDictionary *) args
{
  if (L == NULL)
    {
      if (![self setupLua])
        {
          return NO;
        }
    }


  // flush any global variables besides the tables provided by the libraries
  [self flushGlobals];

  // set up the master coordinate table. Format: rect = {x, y, w, h}
  [self setupLuaRect: r];
  [self pushDict: args];
  lua_setglobal (L, "pave_args");

  // fetch the program's code from the registry
  lua_rawgeti (L, LUA_REGISTRYINDEX, 0);

  if (lua_pcall (L, 0, 0, 0))
    {
      NSLog (@"Error executing PaVe image: %s", lua_tostring (L, -1));
      lua_pop (L, 1);

      return NO;
    }

  return YES;
}

@end

@implementation PaVeImageRep (Private)

// sets up the global 'rect' Lua table
- (void) setupLuaRect: (NSRect) r
{
  lua_newtable (L);
    lua_pushstring (L, "origin");
    lua_newtable (L);
      lua_pushstring (L, "x");
      lua_pushnumber (L, r.origin.x);
      lua_settable (L, -3);

      lua_pushstring (L, "y");
      lua_pushnumber (L, r.origin.y);
      lua_settable (L, -3);
    lua_settable (L, -3);

    lua_pushstring (L, "size");
    lua_newtable (L);
      lua_pushstring (L, "width");
      lua_pushnumber (L, r.size.width);
      lua_settable (L, -3);

      lua_pushstring (L, "height");
      lua_pushnumber (L, r.size.height);
      lua_settable (L, -3);
    lua_settable (L, -3);
  lua_setglobal (L, "rect");
}

// Pushes the Lua equivalent of the given object onto the stack.
// Currently, the only pushable objects are NSString, NSNumber,
// NSArray and NSDictionary
// returns YES if the push succeeded, NO otherwise
- (BOOL) pushObject: obj
{
  lua_checkstack (L, lua_gettop (L) + 1);

  if ([obj isKindOfClass: [NSString class]])
    lua_pushstring (L, [obj cString]);
  else if ([obj isKindOfClass: [NSNumber class]])
    lua_pushnumber (L, [obj doubleValue]);
  else if ([obj isKindOfClass: [NSArray class]])
    [self pushArray: obj];
  else if ([obj isKindOfClass: [NSDictionary class]])
    [self pushDict: obj];
  else
    {
      NSLog (@"Warning: passed invalid argument type %@) to "
        @"PaVe image. Only accepted types are: NSString, NSNumber, "
        @"NSArray and NSDictionary, and their subclasses.", [[obj
        class] name]);

      return NO;
    }

  return YES;
}

// creates a new Lua table which contains all the entries
// from the elements from 'dict'
- (void) pushDict: (NSDictionary *) dict
{
  NSEnumerator * e = [dict keyEnumerator];
  id key;

  lua_newtable (L);

  while ((key = [e nextObject]) != nil)
    {
      id value = [dict objectForKey: key];

      if (![self pushObject: key])
        continue;

      if (![self pushObject: value])
        {
          lua_pop (L, 1);
          continue;
        }

      lua_settable (L, -3);
    }
}

// creates a new Lua table on our global stack which contains
// all the elements from 'array'
- (void) pushArray: (NSArray *) array
{
  lua_newtable (L);

  int i, n, skipped = 0;
  for (i = 0, n = [array count]; i < n; i++)
    {
      id value = [array objectAtIndex: i];

      if (![self pushObject: value])
        {
          skipped++;

          continue;
        }

      lua_rawseti (L, -2, i - skipped);
    }
}

// Prepares a new Lua state and initializes it so as to be usable
// for executing PaVe programs in it
- (BOOL) setupLua
{
  L = lua_open ();
  luaopen_base (L);
  luaopen_table (L);
  luaopen_io (L);
  luaopen_string (L);
  luaopen_math (L);

  // register our operators

#define REG_OP(__op_name__) \
  do { \
    lua_pushcfunction (L, pave_ ## __op_name__); \
    lua_setglobal (L, #__op_name__); \
  } while (0)

  REG_OP (currentalpha);
  REG_OP (currentcmykcolor);
  REG_OP (currentgray);
  REG_OP (currenthsbcolor);
  REG_OP (currentrgbcolor);

  REG_OP (setalpha);
  REG_OP (setgray);
  REG_OP (setrgbcolor);
  REG_OP (sethsbcolor);
  REG_OP (setcmykcolor);

  REG_OP (gsave);
  REG_OP (grestore);

  REG_OP (currentflat);
  REG_OP (currentlinecap);
  REG_OP (currentlinejoin);
  REG_OP (currentlinewidth);
  REG_OP (currentmiterlimit);
  REG_OP (currentpoint);
  REG_OP (currentstrokeadjust);

  REG_OP (setflat);
  REG_OP (sethalftonephase);
  REG_OP (setlinecap);
  REG_OP (setlinejoin);
  REG_OP (setlinewidth);
  REG_OP (setmiterlimit);
  REG_OP (setstrokeadjust);

  REG_OP (concat);
  REG_OP (initmatrix);
  REG_OP (rotate);
  REG_OP (scale);
  REG_OP (translate);

  REG_OP (arc);
  REG_OP (arcn);
  REG_OP (arct);
  REG_OP (clip);
  REG_OP (closepath);
  REG_OP (curveto);
  REG_OP (eoclip);
  REG_OP (eofill);
  REG_OP (fill);
  REG_OP (flattenpath);
  REG_OP (initclip);
  REG_OP (lineto);
  REG_OP (moveto);
  REG_OP (newpath);
  REG_OP (pathbbox);
  REG_OP (rcurveto);
  REG_OP (rectclip);
  REG_OP (rectfill);
  REG_OP (rectstroke);
  REG_OP (reversepath);
  REG_OP (rlineto);
  REG_OP (rmoveto);
  REG_OP (stroke);

  REG_OP (availablefonts);
  REG_OP (availablefontfamilies);
  REG_OP (findfont);
  REG_OP (setfont);
  REG_OP (fontinfo);
  REG_OP (widthofstring);
  REG_OP (show);
  REG_OP (charpath);

  REG_OP (setdash);

#undef REG_OP

  // compile the program
  if (luaL_loadbuffer (L, [programData cString], [programData length], "line"))
    {
      NSLog (@"Error compiling PaVe image %@: %s", name, lua_tostring (L, -1));
      lua_close (L);
      L = NULL;

      return NO;
    }

  // and store it in the registry
  lua_rawseti (L, LUA_REGISTRYINDEX, 0);

  // store a reference to us in the program's registry - some subroutines
  // might need it
  lua_pushstring (L, "imagerep");
  lua_pushlightuserdata (L, self);
  lua_settable (L, LUA_REGISTRYINDEX);

  // remove dangerous IO operations from the environment
  lua_pushnil (L);
  lua_setglobal (L, "io");
  lua_pushnil (L);
  lua_setglobal (L, "dofile");
  lua_pushnil (L);
  lua_setglobal (L, "loadfile");
  lua_pushnil (L);
  lua_setglobal (L, "loadlib");
  lua_pushnil (L);
  lua_setglobal (L, "require");

#define COPY_T(__table__, __name__) \
do { \
  lua_getglobal (L, __table__); \
  lua_pushstring (L, __name__); \
  lua_gettable (L, -2); \
  lua_remove (L, -2); \
  lua_pushstring (L, __name__); \
  lua_insert (L, -2); \
  lua_settable (L, -3); \
} while (0)

  // prepare a new 'os' table and copy allowed entries into it
  // from the old 'os' table
  lua_newtable (L);
  COPY_T ("os", "clock");
  COPY_T ("os", "date");
  COPY_T ("os", "difftime");
  COPY_T ("os", "time");
  lua_setglobal (L, "os");

  return YES;
}

// Removes global symbols which aren't listed in 'permitttedGlobalSymbols'
// from our Lua state. This is used between invocations of our Lua program
// to clear the global variables accumulated by the old runs.
- (void) flushGlobals
{
  // we use this to mark the bottom of our objects to remove
  lua_pushnil (L);

  lua_pushnil (L);  // first key

  while (lua_next (L, LUA_GLOBALSINDEX) != 0)
    {
      // remove the value of the key - we're not interrested in that
      lua_pop (L, 1);

      // check that the symbol is a permitted symbol
      if (!lua_isstring (L, -1) || ![permittedGlobalSymbols containsObject:
        [NSString stringWithCString: lua_tostring (L, -1)]])
        {
          // duplicate the top value, so that later when we walk the stack
          // backwards again, we will remove them
          lua_pushvalue (L, -1);
        }
    }

  while (!lua_isnil (L, -1))
    {
      lua_pushnil (L);
      lua_settable (L, LUA_GLOBALSINDEX);
    }

  // remove the `nil' mark
  lua_pop (L, 1);
}

@end
